----------------------------------------------------------------------------------------------------
----           / --------------
----------------------------------------------------------------------------------------------------
-----------------------------------------------------------------------------
-- |
-- Module      :  ByteStream
-- Copyright   :  (c) Bulat Ziganshin, bulat_z@mail.ru
-- License     :  BSD-style (see the file libraries/base/LICENSE)
--
-- Maintainer  :  bulat_z@mail.ru
-- Stability   :  experimental
-- Portability :  GHC/Hugs on x86 processors
--
--  This module is like 'Binary' module from NHC - it supports writing data
--  structures to binary files or memory buffers and reading them back.
--
--  This module features:
--  * Compatibility with last versions of GHC and Hugs
--  * Lightning speed, especially for large strings and lists of Ints/Words
--      (i have seen a 10mb/s speed on my 1.2 ghz machine)
--  * Flexibility of input/output - data may be hold in files, memory buffers,
--      or reading/writing may be performed via callbacks
--
--  This module currently DON'T supports:
--  * Haskell'98 compatibility (because it uses "too complex" class scheme)
--  * Compatibility with MSB (most significant byte first) processors,
--      including Power PC, Motorola and Sparc
--  * Compatibility with processors, which require aligning of Ints on word
--      boundaries
--  * Bit-oriented compression (instead it uses byte-oriented compression,
--      which is faster and simplier)
--  * Writing strings which contains Unicode or null chars
--  * Tell/Seek-like operations on streams and "freezing" streams
--  * Reading input streams via fixed-size buffer (buffering at this time
--      supported only for output streams, input streams must be placed
--      in one memory buffer containing all the data. MOREOVER, YOU MUST
--      ALLOCATE BUFFER WITH 8 ADDITIONAL BYTES AFTER END OF REAL DATA.
--      It's because Integer demarshalling can pre-read whole 9 bytes
--      even for values which use only 1 byte)
--
--  Example of simple usage you can see in the last section of this file,
--  and examples of defining functions to read/write values of some type -
--  in two preceding sections of file. If you need more explanations -
--  please write me.
--
-----------------------------------------------------------------------------

module ByteStream where

import Prelude hiding (read,readList)
import Control.Exception
import Control.Monad
import Control.Monad.Fix
import Data.Bits
import Data.Char
import Data.IORef
import Data.Word
import Foreign.C.String
import Foreign.C.Types
import Foreign.Marshal.Alloc
import Foreign.Ptr
import Foreign.Storable
import System.IO  hiding (openFile)

aTYPICAL_BUFFER = 64*1024

----------------------------------------------------------------------------------------------------
----        ------------------------------------
----------------------------------------------------------------------------------------------------

data OutStream = OutStream
  { ref_buf     :: IORef (Ptr CChar)           --   ,    
  , ref_size    :: IORef Int                   --    
  , ref_pos     :: IORef Int                   --     
  , functions   :: ( IO (Ptr CChar, Int)              -- ,     
                   , Ptr CChar -> Int -> Int -> IO () -- (.  create)
                   , IO () )
  }

-- |     `filename`,      `size` 
createFile filename size = do
  file <- fileCreate filename
  createBuffered size (\buf size -> fileWriteBuf file buf size) (fileClose file)

-- |  ,      `size` .
-- ,   ,    `writer`
createBuffered size writer closer = do
  buf <- mallocBytes size
  let sendBuf buf size len  =  writer buf len
  create (return (buf,size)) sendBuf (free buf >> closer)

-- |   ,   :
-- receiveBuf : IO (buf,size)                -      
-- sendBuf    : buf -> size -> len -> IO ()  -        `len`
-- cleanup    : IO ()                        - cleanup   
create receiveBuf sendBuf cleanup = do
  (buf, size) <- receiveBuf   --       
  ref_buf  <- newIORef buf
  ref_size <- newIORef size
  ref_pos  <- newIORef 0
  return (OutStream ref_buf ref_size ref_pos (receiveBuf, sendBuf, cleanup))

-- |     
receiveBuffer (OutStream ref_buf ref_size ref_pos (receiveBuf, _, _)) = do
  (buf, size) <- receiveBuf
  writeIORef ref_buf  buf
  writeIORef ref_size size
  writeIORef ref_pos  0

-- |,         `bytes` .
--   -       , ,      !
ensureFreeSpaceInOutStream buffer@(OutStream _ ref_size ref_pos _) bytes = do
  size <- readIORef ref_size
  pos  <- readIORef ref_pos
  when (pos+bytes>size-1) $ do
    sendBuffer buffer
    receiveBuffer buffer
    size <- readIORef ref_size
    pos  <- readIORef ref_pos
    when (pos+bytes>size-1) $
      fail$ "OutStream: needs "++show bytes++" bytes, but entire new buffer contains only "++show size++" bytes"

-- |          
sendBuffer (OutStream ref_buf ref_size ref_pos (_, sendBuf, _)) = do
  modifyIORefIO ref_buf $ \buf -> do
    size <- readIORef ref_size
    pos  <- readIORef ref_pos
    sendBuf buf size pos
    return (error "OutStream::buf undefined")

-- |      
closeOut buffer@(OutStream _ _ _ (_, _, cleanup)) = do
  sendBuffer buffer
  cleanup

-- |All-in-one :   ,       .
--       -    tuple
writeAll receiveBuf sendBuf cleanup value = do
  bracket (create receiveBuf sendBuf cleanup) (closeOut) $  \stream -> do
    write stream value

-- |All-in-one :       .
--       -    tuple
writeFile filename value = do
  bracket (createFile filename aTYPICAL_BUFFER) (closeOut) $  \stream -> do
    write stream value

{-# NOINLINE createFile #-}
{-# NOINLINE createBuffered #-}
{-# NOINLINE create #-}
{-# NOINLINE receiveBuffer #-}
{-# NOINLINE ensureFreeSpaceInOutStream #-}
{-# NOINLINE sendBuffer #-}
{-# NOINLINE closeOut #-}
{-# NOINLINE writeAll #-}


----------------------------------------------------------------------------------------------------
----        ------------------------------------
----------------------------------------------------------------------------------------------------

data InStream = InStream
  { iref_buf     :: IORef (Ptr CChar)           --   ,    
  , iref_size    :: IORef Int                   --    
  , iref_pos     :: IORef Int                   --     
  , ifunctions   :: ( IO (Ptr CChar, Int)              -- ,     
                    , Ptr CChar -> Int -> Int -> IO () -- (.  open)
                    , IO () )
  }

-- |to do:    ,      `size` 
--        
openFile filename _size = do
  file     <- fileOpen filename
  filesize <- hFileSize file             -- temporary solution
  let size  = 8 + fromInteger filesize   -- ditto
  buf      <- mallocBytes size
  let receiveBuf = do len <- fileReadBuf file buf size;  return (buf, len)
      sendBuf buf size len  =  return ()
  open receiveBuf sendBuf (free buf >> fileClose file)

-- | ,    `buf`  `size`
openMemory buf size = do
  ref_bytes_read <- newIORef 0   --      
  let   -- receiveBuf  (buf,size)    ,    
      receiveBuf = do bytes_read <- readIORef ref_bytes_read
                      return (buf `plusPtr` bytes_read, size-bytes_read)
        -- sendBuf ,   `len`   
      sendBuf buf size len  =  modifyIORef ref_bytes_read (len+)
   --   `open`;        
   --    `buf`
  open receiveBuf sendBuf (return ())

-- |to do:    ,   :
-- receiveBuf : IO (buf,size)                -   `buf`    `size`
-- sendBuf    : buf -> size -> len -> IO ()  -   ,    `len` 
-- cleanup    : IO ()                        - cleanup   
open receiveBuf sendBuf cleanup = do
  (buf, size) <- receiveBuf   --       
  ref_buf  <- newIORef buf
  ref_size <- newIORef size
  ref_pos  <- newIORef 0
  return (InStream ref_buf ref_size ref_pos (receiveBuf, sendBuf, cleanup))

-- |      `cleanup`
closeIn (InStream _ _ _ (_, _, cleanup)) = do
  cleanup

-- |All-in-one :   ,     .
--       -    tuple
readMemory buf size = do
  bracket (openMemory buf size) (closeIn) (read)

-- |All-in-one :  ,     .
--       -    tuple
readFile filename = do
  bracket (openFile filename aTYPICAL_BUFFER) (closeIn) (read)

{-# NOINLINE openFile #-}
{-# NOINLINE openMemory #-}
{-# NOINLINE open #-}
{-# NOINLINE closeIn #-}
{-# NOINLINE readMemory #-}



----------------------------------------------------------------------------------------------------
----   ,    /   ----------------------------
----------------------------------------------------------------------------------------------------
-- |           
class BufferData a where
  -- |     
  write :: OutStream -> a -> IO ()

  -- |      -         :)
  writeList :: OutStream -> [a] -> IO ()
  writeList buffer list  =  mapM_ (write buffer) list

  -- |     
  read :: InStream -> IO a

  -- |       -        :)
  readList :: InStream -> Int -> IO [a]
  readList buffer length  =  replicateM length (read buffer)

  {-# NOINLINE read #-}
  {-# NOINLINE write #-}
  {-# NOINLINE readList #-}
  {-# NOINLINE writeList #-}

--      FastBufferData,     BufferData :)
instance (FastBufferData a) => BufferData a where
  write     = writeFast
  writeList = writeListFast
  read      = readFast
  readList  = readListFast


-- |             
class FastBufferData a where
  --       :
  --    - ,      (1  CChar, 4  Int32  ..)
  maxSizeOf :: a -> Int
  --   ,    `buf`   `pos`  `x`,  
  --         ( ,    ,
  --      "pos+maxSizeOf x")
  writeUnchecked :: Ptr CChar -> a -> Int -> IO Int
  --   ,    `buf`   `pos` ,    
  --          
  readUnchecked :: Ptr CChar -> Int -> IO (a, Int)

  -- |     -  
  writeFast :: OutStream -> a -> IO ()
  writeFast buffer@(OutStream ref_buf _ ref_pos _) x = do
    ensureFreeSpaceInOutStream buffer (maxSizeOf x)   -- ,     
    modifyIORefIO ref_pos $ \pos -> do                   --   ref_pos   
      buf <- readIORef ref_buf
      writeUnchecked buf x pos                           --         pos

  -- |-     !
  writeListFast :: OutStream -> [a] -> IO ()
  writeListFast buffer@(OutStream ref_buf _ ref_pos _)   list = do
    let aSIZE = 100
    -- ,       `aSIZE`   
    ensureFreeSpaceInOutStream buffer (aSIZE * maxSizeOf (head list))
    buf <- readIORef ref_buf
    pos <- readIORef ref_pos

    --  "go list pos n"    ,    `pos`,
    --    `list`,    `n` .   
    --  -    `writeListFast`,  ,
    --        100 ,      
    -- ,    
    --
    let go []     pos _  = writeIORef ref_pos pos  --  !       !
        go list   pos 0  = do writeIORef ref_pos pos     --     
                              writeListFast buffer list  -- ...       
        go (x:xs) pos n  = do new_pos <- writeUnchecked buf x pos    --   
                              go xs new_pos (n-1)                    -- ...    
    go list pos aSIZE  --      ,    `aSIZE` 


  -- |      
  readFast :: InStream -> IO a
  readFast buffer@(InStream ref_buf _ ref_pos _) = do
    buf <- readIORef ref_buf
    pos <- readIORef ref_pos
    (x, new_pos) <- readUnchecked buf pos
    writeIORef ref_pos new_pos
    return x

  -- |      
  readListFast :: InStream -> Int -> IO [a]
  readListFast buffer@(InStream ref_buf _ ref_pos _) length = do
    buf <- readIORef ref_buf
    pos <- readIORef ref_pos
    let go pos 0 xs = do writeIORef ref_pos pos
                         return (reverse xs)
        go pos n xs = do (x, new_pos) <- readUnchecked buf pos
                         go new_pos (n-1) (x:xs)
    go pos length []

  {-# NOINLINE readFast #-}
  {-# NOINLINE writeFast #-}
  {-# NOINLINE readListFast #-}
  {-# NOINLINE writeListFast #-}



----------------------------------------------------------------------------------------------------
----      -----------------------------------------------------------
----------------------------------------------------------------------------------------------------

--    Storable     FastBufferData:
--  ,      ,      /  
instance (Storable a) => FastBufferData a where
  maxSizeOf x = sizeOf x
  writeUnchecked buf x pos = do
    pokeByteOff buf pos x
    return (pos + sizeOf x)
  readUnchecked buf pos = do
    x <- peekByteOff buf pos
    return (x, pos + sizeOf x)


--    CChar,           :(
instance FastBufferData Char where
  maxSizeOf x = maxSizeOf (undefined :: CChar)
  writeUnchecked buf x pos  =  writeUnchecked buf (castCharToCChar x) pos
  readUnchecked buf pos  =  do (x,new_pos) <- readUnchecked buf pos; return (castCCharToChar x, new_pos)


--      ,       (  )
instance BufferData String where
  write buf str  =  writeList buf str  >>  write buf '\0'

  read  buffer@(InStream ref_buf _ ref_pos _) = do
    buf <- readIORef ref_buf
    let go xs pos = do (x,new_pos) <- readUnchecked buf pos
                       case (x) of
                         '\0' -> writeIORef ref_pos new_pos  >>  return (reverse xs)
                         _    -> go (x:xs) new_pos
    go ""  =<<  readIORef ref_pos


--      ,     7-zip.
--      2^64      :  1  9
instance FastBufferData Integer where
  maxSizeOf x = 9   -- : 1     8  
  writeUnchecked buf x pos = do
    let write1  x pos  =  writeUnchecked buf (x::Word8)  pos
        write4  x pos  =  writeUnchecked buf (x::Word32) pos
        write_8 x pos  =  writeUnchecked buf (x::Word64) pos
    --     4  8 ,       
    --  . -  -     , 
    --    ,    
    --      ,        !!!
    --  ,    32- ,  64-    
    case () of
     _ | x<0       ->  fail$ "Sorry, FastBufferData.Integer.writeUnchecked don't support negative values like this: "++show x
       | x<128     ->  do write4  (fromInteger x *  2     ) pos; return (pos+1)
       | x<128^2   ->  do write4  (fromInteger x *  4 +  1) pos; return (pos+2)
       | x<128^3   ->  do write4  (fromInteger x *  8 +  3) pos; return (pos+3)
       | x<128^4   ->  do write4  (fromInteger x * 16 +  7) pos; return (pos+4)
       | x<128^5   ->  do write_8 (fromInteger x * 32 + 15) pos; return (pos+5)
       | x<128^6   ->  do write_8 (fromInteger x * 64 + 31) pos; return (pos+6)
       | x<128^7   ->  do write_8 (fromInteger x *128 + 63) pos; return (pos+7)
       | x<128^8   ->  do write_8 (fromInteger x *256 +127) pos; return (pos+8)
       | x<256^8   ->  do write1 255 pos  >>=  write_8 (fromInteger x); return (pos+9)
       | otherwise ->  fail$ "Sorry, FastBufferData.Integer.writeUnchecked don't support numbers larger than 256^8, like this: "++show x

  readUnchecked buf pos = do
    --     4 ,       ,  
    (x::Word32,_)  <-  readUnchecked buf pos
    case () of
     _ | x .&.  1 ==   0  ->  return (toInteger$ (x `mod` 256^1) `shiftR` 1, pos+1)
       | x .&.  3 ==   1  ->  return (toInteger$ (x `mod` 256^2) `shiftR` 2, pos+2)
       | x .&.  7 ==   3  ->  return (toInteger$ (x `mod` 256^3) `shiftR` 3, pos+3)
       | x .&. 15 ==   7  ->  return (toInteger$ (x            ) `shiftR` 4, pos+4)
       | otherwise -> do
          --     4- ,     8      
          (x::Word64,_)  <-  readUnchecked buf pos
          case () of
           _ | x .&. 31 ==  15  ->  return (toInteger$ (x `mod` 256^5) `shiftR` 5, pos+5)
             | x .&. 63 ==  31  ->  return (toInteger$ (x `mod` 256^6) `shiftR` 6, pos+6)
             | x .&.127 ==  63  ->  return (toInteger$ (x `mod` 256^7) `shiftR` 7, pos+7)
             | x .&.255 == 127  ->  return (toInteger$ (x            ) `shiftR` 8, pos+8)
             | otherwise        ->  do
                 --    -      8   
                 (x::Word64, _) <- readUnchecked buf (pos+1); return (toInteger x, pos+9)


--        Word8 (.  
--   ),           
instance FastBufferData Bool where
  maxSizeOf x = maxSizeOf (undefined :: Word8)
  writeUnchecked buf x pos  =  writeUnchecked buf (toEnum (fromEnum x) :: Word8) pos
  readUnchecked buf pos  =  do (x::Word8, new_pos) <- readUnchecked buf pos; return (toEnum (fromEnum x), new_pos)
{-
  writeListFast buffer list  =  writeListFast buffer (makeBytes list :: [Word8])
   where
    makeBytes (a:b:c:d:e:f:g:h:xs) = (((((((n a*2+n b)*2+n c)*2+n d)*2+n e)*2+n f)*2+n g)*2+n h) : makeBytes xs
    makeBytes [a,b,c,d,e,f,g]      = (((((((n a*2+n b)*2+n c)*2+n d)*2+n e)*2+n f)*2+n g)*2) : []
    makeBytes [a,b,c,d,e,f]        = (((((((n a*2+n b)*2+n c)*2+n d)*2+n e)*2+n f)*2)*2) : []
    makeBytes [a,b,c,d,e]          = (((((((n a*2+n b)*2+n c)*2+n d)*2+n e)*2)*2)*2) : []
    makeBytes [a,b,c,d]            = (((((((n a*2+n b)*2+n c)*2+n d)*2)*2)*2)*2) : []
    makeBytes [a,b,c]              = (((((((n a*2+n b)*2+n c)*2)*2)*2)*2)*2) : []
    makeBytes [a,b]                = (((((((n a*2+n b)*2)*2)*2)*2)*2)*2) : []
    makeBytes [a]                  = (((((((n a*2)*2)*2)*2)*2)*2)*2) : []
    makeBytes []                   = []
    n x = toEnum (fromEnum x);  n :: Bool->Word8
-}



----------------------------------------------------------------------------------------------------
----      ---------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- |,           
writeInteger buf  =  write buf.toInteger
readInteger  buf  =  return.fromInteger =<< read buf

instance (BufferData a) => BufferData [a]  where
  write buf list =  writeInteger buf (length list)  >>   writeList buf list
  read  buf      =  readInteger  buf                >>=  readList  buf

instance (BufferData a, BufferData b) => BufferData (a,b)  where
  write buf (a,b) = write buf a  >>  write buf b
  read  buf       = do a <- read buf; b <- read buf; return (a,b)

instance (BufferData a, BufferData b, BufferData c) => BufferData (a,b,c)  where
  write buf (a,b,c) = write buf ((a,b),c)
  read  buf         = do ((a,b),c) <- read buf; return (a,b,c)

instance (BufferData a, BufferData b, BufferData c, BufferData d) => BufferData (a,b,c,d)  where
  write buf (a,b,c,d) = write buf ((a,b),c,d)
  read  buf           = do ((a,b),c,d) <- read buf; return (a,b,c,d)

instance (BufferData a, BufferData b, BufferData c, BufferData d, BufferData e) => BufferData (a,b,c,d,e)  where
  write buf (a,b,c,d,e) = write buf ((a,b),c,d,e)
  read  buf             = do ((a,b),c,d,e) <- read buf; return (a,b,c,d,e)

instance (BufferData a, BufferData b, BufferData c, BufferData d, BufferData e, BufferData f) => BufferData (a,b,c,d,e,f)  where
  write buf (a,b,c,d,e,f) = write buf ((a,b),c,d,e,f)
  read  buf               = do ((a,b),c,d,e,f) <- read buf; return (a,b,c,d,e,f)

instance (BufferData a, BufferData b, BufferData c, BufferData d, BufferData e, BufferData f, BufferData g) => BufferData (a,b,c,d,e,f,g)  where
  write buf (a,b,c,d,e,f,g) = write buf ((a,b),c,d,e,f,g)
  read  buf                 = do ((a,b),c,d,e,f,g) <- read buf; return (a,b,c,d,e,f,g)

instance (BufferData a, BufferData b, BufferData c, BufferData d, BufferData e, BufferData f, BufferData g, BufferData h) => BufferData (a,b,c,d,e,f,g,h)  where
  write buf (a,b,c,d,e,f,g,h) = write buf ((a,b),c,d,e,f,g,h)
  read  buf                   = do ((a,b),c,d,e,f,g,h) <- read buf; return (a,b,c,d,e,f,g,h)

instance (BufferData a) => BufferData (Maybe a)  where
  write buf (Just  a) = write buf (True,a)
  write buf (Nothing) = write buf False
  read buf = do x <- read buf; if x  then (return.Just =<< read buf)  else (return Nothing)

instance (BufferData a, BufferData b) => BufferData (Either a b)  where
  write buf (Left  a) = write buf (True, a)
  write buf (Right b) = write buf (False,b)
  read buf = do x <- read buf; if x  then (return.Left =<< read buf)  else (return.Right =<< read buf)

{-      / 
class DerivedBufferData a where
  toTuple   :: BufferData b => a -> b
  fromTuple :: BufferData b => b -> a

instance DerivedBufferData a => BufferData a where
  write buf a  =  write buf (toTuple a)
  read  buf    =  do a <- read buf; return (fromTuple a)

instance (BufferData a, BufferData b) => DerivedBufferData (Either a b)  where
  toTuple   (Left  a)  = (True,  a)
  toTuple   (Right b)  = (False, b)
  fromTuple (True,  a) = (Left   a)
  fromTuple (False, b) = (Right  b)

instance (Enum a) => BufferData a  where
  write buf a = writeInteger buf (fromEnum a)
  read  buf   = readInteger  buf >>= return.toEnum
-}

instance (FastBufferData a, FastBufferData b) => FastBufferData (a,b)  where
  maxSizeOf (a,b)               =  maxSizeOf a + maxSizeOf b
  writeUnchecked buf (a,b) pos  =  writeUnchecked buf a pos  >>=  writeUnchecked buf b
  readUnchecked buf pos         =  do
    (a, pos) <- readUnchecked buf pos
    (b, pos) <- readUnchecked buf pos
    return ((a,b), pos)
  {-# NOINLINE writeUnchecked #-}
  {-# NOINLINE readUnchecked #-}

instance (FastBufferData a, FastBufferData b, FastBufferData c) => FastBufferData (a,b,c)  where
  maxSizeOf (a,b,c)               =  maxSizeOf a + maxSizeOf b + maxSizeOf c
  writeUnchecked buf (a,b,c) pos  =
    writeUnchecked buf a pos
      >>=  writeUnchecked buf b
      >>=  writeUnchecked buf c
  readUnchecked buf pos           =  do
    (a, pos) <- readUnchecked buf pos
    (b, pos) <- readUnchecked buf pos
    (c, pos) <- readUnchecked buf pos
    return ((a,b,c), pos)
  {-# NOINLINE writeUnchecked #-}
  {-# NOINLINE readUnchecked #-}

instance (FastBufferData a, FastBufferData b, FastBufferData c, FastBufferData d) => FastBufferData (a,b,c,d)  where
  maxSizeOf (a,b,c,d)               =  maxSizeOf a + maxSizeOf b + maxSizeOf c + maxSizeOf d
  writeUnchecked buf (a,b,c,d) pos  =
    writeUnchecked buf a pos
      >>=  writeUnchecked buf b
      >>=  writeUnchecked buf c
      >>=  writeUnchecked buf d
  readUnchecked buf pos             =  do
    (a, pos) <- readUnchecked buf pos
    (b, pos) <- readUnchecked buf pos
    (c, pos) <- readUnchecked buf pos
    (d, pos) <- readUnchecked buf pos
    return ((a,b,c,d), pos)
  {-# NOINLINE writeUnchecked #-}
  {-# NOINLINE readUnchecked #-}



----------------------------------------------------------------------------------------------------
----   -----------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- | ,    IORef,  ,
--       ,   
modifyIORefIO :: IORef a -> (a -> IO a) -> IO ()
modifyIORefIO var action = do
  readIORef var  >>=  action  >>=  writeIORef var


-- |   
type File            = Handle
fileGetPos           = hTell
fileGetSize          = hFileSize
fileSeek   file      = hSeek file AbsoluteSeek
fileRead   file size = replicateM size (hGetChar file)
fileWrite            = hPutStr
fileCreate name      = openBinaryFile name WriteMode
fileOpen   name      = openBinaryFile name ReadMode
fileGet    name      = hGetContents =<< fileOpen name
fileClose  file      = hClose file

fileReadAt file pos size = do
  fileSeek file pos
  fileRead file size

filePut name data1 = do
  file <- fileCreate name
  fileWrite file data1
  fileClose file

#if defined(__GLASGOW_HASKELL__)

-- Read/write to raw memory buffer
fileReadBuf   =  hGetBuf
fileWriteBuf  =  hPutBuf

#else

fileReadBuf file buf (size::Int) = do
  filesize <- fileGetSize file
  filepos  <- fileGetPos file
  x <- fileRead file (fromInteger$ min (toInteger size) (filesize-filepos))
  let len = length x
  pokeCStringLen (buf,len) x
  return len

fileWriteBuf file buf (size::Int) = do
  fileWrite file  =<<  peekCStringLen (buf, size)

#endif



----------------------------------------------------------------------------------------------------
---- Example of simple usage of in/out byte streams ------------------------------------------------
----------------------------------------------------------------------------------------------------
{-
test = do
  stream <- ByteStream.createFile "test" 5000
  ByteStream.write stream  "asdfr"
  ByteStream.write stream  "12345"
  ByteStream.write stream  [10,20..500::Int]
  ByteStream.write stream  ([10,20..500] ++ [103*10^3, 106*10^6, 109*10^9, 112*10^12, 115*10^15::Integer])
  ByteStream.write stream  (concat$ replicate 100 [True,False,True])
  ByteStream.closeOut stream

  stream <- ByteStream.openFile "test" 5000
  (x::String)    <- ByteStream.read stream
  (y::String)    <- ByteStream.read stream
  (a::[Int])     <- ByteStream.read stream
  (b::[Integer]) <- ByteStream.read stream
  (c::[Bool])    <- ByteStream.read stream
  ByteStream.closeIn stream
  print [x,y]
  print a
  print b
  print c
-}
--Checklist:
--1. +   receiveBuf = receiveP pipe
--2. +write  Storable   -       pokeByteOff elem
--3. +  
--4. + writeUnchecked
--5. +  Integer
--6. + WriteList  WriteListWithoutLength
--7. +        read/writeList=writeLength+writeListWithoutLength
--8.       
--9.   100   
--10.   [Bool]

